from typing import Any, Optional, Union

import pydantic

from ...sentinels import NotSet, NotSetType
from ...updates import MetadataChangeset
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionReference,
)


class CreateActionRequest(pydantic.BaseModel):
    # Required
    name: str

    # Optional
    parameters: Optional[list[ActionParameter]] = pydantic.Field(default_factory=list)
    description: Optional[str] = None
    uri: Optional[str] = None
    inherits: Optional[ActionReference] = None
    metadata: Optional[dict[str, Any]] = pydantic.Field(default_factory=dict)
    tags: Optional[list[str]] = pydantic.Field(default_factory=list)
    compute_requirements: Optional[ComputeRequirements] = None
    container_parameters: Optional[ContainerParameters] = None


class QueryActionsRequest(pydantic.BaseModel):
    filters: dict[str, Any] = pydantic.Field(default_factory=dict)

    class Config:
        extra = "forbid"


class SetActionAccessibilityRequest(pydantic.BaseModel):
    accessibility: Accessibility
    digest: Optional[str] = None
    """Specify specific version of Action. If not specified, the latest version's accessibility will be updated."""

    class Config:
        extra = "forbid"


class UpdateActionRequest(pydantic.BaseModel):
    compute_requirements: Union[ComputeRequirements, NotSetType] = NotSet
    container_parameters: Union[ContainerParameters, NotSetType] = NotSet
    description: Optional[Union[str, NotSetType]] = NotSet
    metadata_changeset: Union[MetadataChangeset, NotSetType] = NotSet
    parameter_changeset: Union[ActionParameterChangeset, NotSetType] = NotSet
    uri: Union[str, NotSetType] = NotSet

    @pydantic.validator("uri")
    def validate_uri(cls, v):
        if v is NotSet:
            return v

        stripped = v.strip()
        if not stripped:
            raise ValueError("uri cannot be empty")
        return stripped

    class Config:
        extra = "forbid"
        schema_extra = NotSetType.openapi_schema_modifier



from typing import Any, Optional
import urllib.parse

from ...exceptions import RobotoHttpExceptionParse
from ...http import (
    HttpClient,
    PaginatedList,
    StreamedList,
    roboto_headers,
)
from ...logging import default_logger
from ...query import QuerySpecification
from ...serde import pydantic_jsonable_dict
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .action_record import ActionRecord
from .invocation_delegate import (
    InvocationDelegate,
)
from .invocation_http_resources import (
    CreateInvocationRequest,
    SetLogsLocationRequest,
)
from .invocation_record import (
    InvocationDataSourceType,
    InvocationRecord,
    InvocationSource,
    InvocationStatus,
    LogRecord,
    LogsLocation,
)

logger = default_logger()


class InvocationHttpDelegate(InvocationDelegate):
    __http_client: HttpClient
    __roboto_service_base_url: str

    def __init__(self, roboto_service_base_url: str, http_client: HttpClient) -> None:
        super().__init__()
        self.__http_client = http_client
        self.__roboto_service_base_url = roboto_service_base_url

    def headers(
        self,
        org_id: Optional[str] = None,
        user_id: Optional[str] = None,
        resource_owner_id: Optional[str] = None,
    ) -> dict[str, str]:
        return roboto_headers(
            org_id=org_id,
            user_id=user_id,
            resource_owner_id=resource_owner_id,
            additional_headers={"Content-Type": "application/json"},
        )

    def create_invocation(
        self,
        action_record: ActionRecord,
        parameter_values: dict[str, Any],
        input_data: list[str],
        data_source_id: str,
        data_source_type: InvocationDataSourceType,
        invocation_source: InvocationSource,
        invocation_source_id: Optional[str] = None,
        compute_requirement_overrides: Optional[ComputeRequirements] = None,
        container_parameter_overrides: Optional[ContainerParameters] = None,
        idempotency_id: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> InvocationRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions/{action_record.name}/invoke?digest={action_record.digest}"

        request_body = CreateInvocationRequest(
            parameter_values=parameter_values,
            input_data=input_data,
            data_source_id=data_source_id,
            data_source_type=data_source_type,
            invocation_source=invocation_source,
            invocation_source_id=invocation_source_id,
            idempotency_id=idempotency_id,
            compute_requirement_overrides=compute_requirement_overrides,
            container_parameter_overrides=container_parameter_overrides,
        )

        user_id = (
            invocation_source_id
            if invocation_source == InvocationSource.Manual
            else None
        )
        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url,
                data=pydantic_jsonable_dict(request_body, exclude_none=True),
                headers=self.headers(
                    org_id=org_id,
                    user_id=user_id,
                    resource_owner_id=action_record.org_id,
                ),
            )

        return InvocationRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_by_id(
        self, invocation_id: str, org_id: Optional[str] = None
    ) -> InvocationRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions/invocations/{invocation_id}"
        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url, headers=self.headers(org_id))

        return InvocationRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_logs(
        self,
        invocation_id: str,
        page_token: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> PaginatedList[LogRecord]:
        url = f"{self.__roboto_service_base_url}/v1/actions/invocations/{invocation_id}/logs"
        if page_token:
            encoded_qs = urllib.parse.urlencode({"page_token": page_token})
            url = f"{url}?{encoded_qs}"

        with RobotoHttpExceptionParse():
            http_response = self.__http_client.get(url, self.headers(org_id))
        data = http_response.from_json(json_path=["data"])

        return PaginatedList(
            items=[LogRecord.parse_obj(record) for record in data["items"]],
            next_token=data["next_token"],
        )

    def stream_logs(
        self,
        invocation_id: str,
        last_read: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> StreamedList[LogRecord]:
        url = f"{self.__roboto_service_base_url}/v1/actions/invocations/{invocation_id}/logs/stream"
        if last_read:
            encoded_qs = urllib.parse.urlencode({"last_read": str(last_read)})
            url = f"{url}?{encoded_qs}"

        with RobotoHttpExceptionParse():
            http_response = self.__http_client.get(url, self.headers(org_id))
        data = http_response.from_json(json_path=["data"])

        return StreamedList(
            items=[LogRecord.parse_obj(record) for record in data["items"]],
            has_next=data["has_next"],
            last_read=data["last_read"],
        )

    def set_logs_location(
        self,
        record: InvocationRecord,
        logs: LogsLocation,
    ) -> InvocationRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions/invocations/{record.invocation_id}/logs"
        request_body = SetLogsLocationRequest(bucket=logs.bucket, prefix=logs.prefix)

        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url,
                data=pydantic_jsonable_dict(request_body, exclude_none=True),
                headers=self.headers(record.org_id),
            )
        return InvocationRecord.parse_obj(response.from_json(json_path=["data"]))

    def update_invocation_status(
        self,
        record: InvocationRecord,
        status: InvocationStatus,
        detail: Optional[str] = None,
    ) -> InvocationRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions/invocations/{record.invocation_id}/status"

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url,
                data={"status": status.value, "detail": detail},
                headers=self.headers(record.org_id),
            )
        return InvocationRecord.parse_obj(response.from_json(json_path=["data"]))

    def query_invocations(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[InvocationRecord]:
        url = f"{self.__roboto_service_base_url}/v1/actions/invocations/query"
        post_body = pydantic_jsonable_dict(query, exclude_none=True)
        with RobotoHttpExceptionParse():
            res = self.__http_client.post(
                url,
                data=post_body,
                headers=self.headers(org_id),
            )

        unmarshalled = res.from_json(json_path=["data"])
        return PaginatedList(
            items=[
                InvocationRecord.parse_obj(record) for record in unmarshalled["items"]
            ],
            next_token=unmarshalled["next_token"],
        )



from typing import Any, Optional

import pydantic

from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .invocation_record import (
    InvocationDataSourceType,
    InvocationSource,
    InvocationStatus,
)


class CreateInvocationRequest(pydantic.BaseModel):
    parameter_values: Optional[dict[str, Any]]
    input_data: list[str]
    compute_requirement_overrides: Optional[ComputeRequirements]
    container_parameter_overrides: Optional[ContainerParameters]
    data_source_id: str
    data_source_type: InvocationDataSourceType
    invocation_source: InvocationSource
    invocation_source_id: Optional[str]
    idempotency_id: Optional[str]


class QueryInvocationsRequest(pydantic.BaseModel):
    filters: dict[str, Any] = pydantic.Field(default_factory=dict)

    class Config:
        extra = "forbid"


class SetLogsLocationRequest(pydantic.BaseModel):
    bucket: str
    prefix: str


class UpdateInvocationStatus(pydantic.BaseModel):
    status: InvocationStatus
    detail: str



import enum
import typing

import pydantic

from .conditions import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
)


class SortDirection(str, enum.Enum):
    """The direction to sort the results of a query."""

    Ascending = "ASC"
    Descending = "DESC"


class QuerySpecification(pydantic.BaseModel):
    """
    Model for specifying a query to the Roboto Platform.

    Examples:
        Specify a query with a single condition:
            >>> from roboto import query
            >>> query_spec = query.QuerySpecification(
            ...     condition=query.Condition(
            ...         field="name",
            ...         comparator=query.Comparator.Equals,
            ...         value="Roboto"
            ...     )
            ... )

        Specify a query with multiple conditions:
            >>> from roboto import query
            >>> query_spec = query.QuerySpecification(
            ...     condition=query.ConditionGroup(
            ...         operator=query.ConditionOperator.And,
            ...         conditions=[
            ...             query.Condition(
            ...                 field="name",
            ...                 comparator=query.Comparator.Equals,
            ...                 value="Roboto"
            ...             ),
            ...             query.Condition(
            ...                 field="age",
            ...                 comparator=query.Comparator.GreaterThan,
            ...                 value=18
            ...             )
            ...         ]
            ...     )
            ... )

        Arbitrarily nest condition groups:
            >>> from roboto import query
            >>> query_spec = query.QuerySpecification(
            ...     condition=query.ConditionGroup(
            ...         operator=query.ConditionOperator.And,
            ...         conditions=[
            ...             query.Condition(
            ...                 field="name",
            ...                 comparator=query.Comparator.Equals,
            ...                 value="Roboto"
            ...             ),
            ...             query.ConditionGroup(
            ...                 operator=query.ConditionOperator.Or,
            ...                 conditions=[
            ...                     query.Condition(
            ...                         field="age",
            ...                         comparator=query.Comparator.GreaterThan,
            ...                         value=18
            ...                     ),
            ...                     query.Condition(
            ...                         field="age",
            ...                         comparator=query.Comparator.LessThan,
            ...                         value=30
            ...                     )
            ...                 ]
            ...             )
            ...         ]
            ...     )
            ... )
    """

    condition: typing.Optional[typing.Union[Condition, ConditionGroup]] = None
    limit: int = 1000
    after: typing.Optional[str] = None  # An encoded PaginationToken
    sort_by: typing.Optional[str] = None
    sort_direction: typing.Optional[SortDirection] = None

    class Config:
        extra = "forbid"

    @classmethod
    def protected_from_oldstyle_request(
        cls, filters: dict, page_token: typing.Optional[str] = None
    ) -> "QuerySpecification":
        """
        Convert deprecated query format to new format.

        Not for public use.

        :meta private:
        """
        conditions: list[typing.Union[Condition, ConditionGroup]] = []

        def _iterconditions(field: str, value: typing.Any):
            if isinstance(value, list):
                conditions.append(
                    ConditionGroup(
                        operator=ConditionOperator.Or,
                        conditions=[
                            Condition(
                                field=field, comparator=Comparator.Equals, value=v
                            )
                            for v in value
                        ],
                    )
                )

            elif isinstance(value, dict):
                for k, v in value.items():
                    _iterconditions(f"{field}.{k}", v)

            else:
                conditions.append(
                    Condition(field=field, comparator=Comparator.Equals, value=value)
                )

        for field, value in filters.items():
            _iterconditions(field, value)

        condition: typing.Optional[ConditionGroup] = None
        if len(conditions) > 0:
            condition = ConditionGroup(
                operator=ConditionOperator.And, conditions=conditions
            )

        return cls(
            condition=condition,
            after=page_token,
        )

    def fields(self) -> set[str]:
        """Return a set of all fields referenced in the query."""
        fields = set()

        def _iterconditions(
            condition: typing.Optional[typing.Union[Condition, ConditionGroup]]
        ):
            if condition is None:
                return

            if isinstance(condition, Condition):
                fields.add(condition.field)
            else:
                for cond in condition.conditions:
                    _iterconditions(cond)

        _iterconditions(self.condition)
        return fields



#  Copyright (c) 2023 Roboto Technologies, Inc.

import argparse
import os

ORG_ARG_HELP = (
    "The org_id for the calling organization. Will be set implicitly "
    + "if the user is part of exactly one org. "
    + "Users can set a `ROBOTO_ORG_ID` environment variable to control the default value of this argument."
)

DEFAULT_ORG_ID = os.getenv("ROBOTO_ORG_ID")


def add_org_arg(parser: argparse.ArgumentParser, arg_help: str = ORG_ARG_HELP):
    parser.add_argument(
        "--org", required=False, type=str, help=arg_help, default=DEFAULT_ORG_ID
    )



import datetime
import enum
import hashlib
import json
import typing

import pydantic

from ...serde import pydantic_jsonable_dict
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)


class Accessibility(str, enum.Enum):
    """
    Controls who can query for and invoke an action.

    Future accessibility levels may include: "user" and/or "team".
    """

    Organization = "organization"
    """All members of the organization owning the Action can query for and invoke the action."""

    ActionHub = "action_hub"
    """All users of Roboto can query for and invoke the action."""


class ActionParameter(pydantic.BaseModel):
    name: str
    required: bool = False
    description: typing.Optional[str] = None
    default: typing.Optional[typing.Any] = None

    class Config:
        extra = "forbid"


class ActionParameterChangeset(pydantic.BaseModel):
    put_parameters: list[ActionParameter] = pydantic.Field(default_factory=list)
    remove_parameters: list[str] = pydantic.Field(default_factory=list)

    def is_empty(self) -> bool:
        return not self.put_parameters and not self.remove_parameters


class ActionReference(pydantic.BaseModel):
    """Qualified action reference."""

    name: str
    digest: typing.Optional[str] = None
    owner: typing.Optional[str] = None


class ActionRecord(pydantic.BaseModel):
    # Required fields without defaults
    created: datetime.datetime  # Persisted as ISO 8601 string in UTC
    created_by: str
    modified: datetime.datetime  # Persisted as ISO 8601 string in UTC
    modified_by: str
    name: str  # Sort key
    org_id: str  # Partition key

    # Optional fields with defaults
    accessibility: Accessibility = Accessibility.Organization
    compute_requirements: typing.Optional[ComputeRequirements] = None
    container_parameters: typing.Optional[ContainerParameters] = None
    description: typing.Optional[str] = None
    digest: typing.Optional[str] = None
    inherits: typing.Optional[ActionReference] = None
    metadata: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    parameters: list[ActionParameter] = pydantic.Field(default_factory=list)
    # Persisted as ISO 8601 string in UTC
    published: typing.Optional[datetime.datetime] = None
    tags: list[str] = pydantic.Field(default_factory=list)
    uri: typing.Optional[str] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.digest = self.compute_digest() if not self.digest else self.digest

    def compute_digest(self) -> str:
        hasher = hashlib.blake2b(
            digest_size=16,
            # https://docs.python.org/3.9/library/hashlib.html#personalization
            person=b"ActionRecord",
        )
        digestable = pydantic_jsonable_dict(self, exclude_unset=True)
        hasher.update(json.dumps(digestable, sort_keys=True).encode("utf-8"))
        return hasher.hexdigest()

    @property
    def reference(self) -> ActionReference:
        return ActionReference(
            name=self.name,
            digest=self.digest,
            owner=self.org_id,
        )



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ...serde import pydantic_jsonable_dict
from .delegate import OrgDelegate
from .record import (
    OrgRecord,
    OrgRoleName,
    OrgStatus,
)


class Org:
    __record: OrgRecord
    __org_delegate: OrgDelegate

    @classmethod
    def create(
        cls,
        name: str,
        org_delegate: OrgDelegate,
        bind_email_domain: bool = False,
        creator_user_id: Optional[str] = None,
    ):
        record = org_delegate.create_org(
            creator_user_id=creator_user_id,
            name=name,
            bind_email_domain=bind_email_domain,
        )
        return cls(record=record, org_delegate=org_delegate)

    @classmethod
    def from_id(cls, org_id: str, org_delegate: OrgDelegate) -> "Org":
        record = org_delegate.get_org_by_id(org_id=org_id)
        return cls(record=record, org_delegate=org_delegate)

    @classmethod
    def for_user(
        cls, org_delegate: OrgDelegate, user_id: Optional[str] = None
    ) -> list["Org"]:
        records = org_delegate.orgs_for_user(user_id=user_id)
        return [cls(record=record, org_delegate=org_delegate) for record in records]

    @property
    def record(self) -> OrgRecord:
        return self.__record

    def update(
        self, updates: dict[str, Any], caller_user_id: Optional[str] = None
    ) -> "Org":
        self.__record = self.__org_delegate.update_org(
            updates=updates, org_id=self.org_id, caller_user_id=caller_user_id
        )
        return self

    def delete(self) -> None:
        self.__org_delegate.delete_org(org_id=self.org_id)

    def bind_email_domain(self, email_domain: str) -> None:
        self.__org_delegate.bind_email_domain(
            org_id=self.__record.org_id, email_domain=email_domain
        )

    def unbind_email_domain(self, email_domain: str) -> None:
        self.__org_delegate.unbind_email_domain(
            org_id=self.__record.org_id, email_domain=email_domain
        )

    def list_email_domains(self) -> list[str]:
        return self.__org_delegate.get_email_domains_for_org(
            org_id=self.__record.org_id
        )

    def remove_user(self, user_id: str) -> None:
        """
        Remove a user from an organization.

        Args:
            user_id: The user_id of the user to remove from this organization.

        Raises:
            roboto_sdk.exceptions.RobotoUnauthorizedException: In a number of different cases:
                - The caller is not a member of this org
                - The caller is not an org administrator or owner of this org
                - The caller is an administrator but not an owner, and the target user is also an administrator
            roboto_sdk.exceptions.RobotoNotFoundException: If the target user is not a member of the specified org.
            roboto_sdk.exceptions.RobotoInvalidRequestException:
                If the caller is trying to remove themselves from the org, and the caller is the owner of that org.
                The admin should transfer ownership of the org to another user before removing themselves.

        Returns:
            None
        """
        self.__org_delegate.remove_user_from_org(user_id=user_id, org_id=self.org_id)

    def add_role_for_user(self, user_id: str, role_name: OrgRoleName):
        self.__org_delegate.add_role_for_user(
            user_id=user_id, role_name=role_name, org_id=self.org_id
        )

    def remove_role_from_user(self, user_id: str, role_name: OrgRoleName):
        self.__org_delegate.remove_role_from_user(
            user_id=user_id, role_name=role_name, org_id=self.org_id
        )

    def __init__(self, record: OrgRecord, org_delegate: OrgDelegate):
        self.__record = record
        self.__org_delegate = org_delegate

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)

    @property
    def org_id(self):
        return self.__record.org_id

    @property
    def status(self) -> OrgStatus:
        return self.__record.status



#  Copyright (c) 2023 Roboto Technologies, Inc.
import typing

import pydantic

from .record import OrgRoleName, OrgStatus


class CreateOrgRequest(pydantic.BaseModel):
    name: str
    bind_email_domain: bool = False


class OrgRecordUpdates(pydantic.BaseModel):
    name: typing.Optional[str] = None
    status: typing.Optional[OrgStatus] = None


class UpdateOrgRequest(pydantic.BaseModel):
    updates: OrgRecordUpdates


class BindEmailDomainRequest(pydantic.BaseModel):
    email_domain: str


class InviteUserRequest(pydantic.BaseModel):
    invited_user_id: str


class ModifyRoleForUserRequest(pydantic.BaseModel):
    user_id: str
    role_name: OrgRoleName


class RemoveUserFromOrgRequest(pydantic.BaseModel):
    user_id: str



#  Copyright (c) 2023 Roboto Technologies, Inc.
from .dataset import Dataset
from .delegate import Credentials, DatasetDelegate
from .http_delegate import DatasetHttpDelegate
from .http_resources import (
    CreateDatasetRequest,
    QueryDatasetsRequest,
    UpdateDatasetRequest,
)
from .record import (
    Administrator,
    DatasetRecord,
    StorageLocation,
)

__all__ = (
    "Administrator",
    "CreateDatasetRequest",
    "Credentials",
    "Dataset",
    "DatasetDelegate",
    "DatasetHttpDelegate",
    "DatasetRecord",
    "QueryDatasetsRequest",
    "StorageLocation",
    "UpdateDatasetRequest",
)



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import shlex
import subprocess
import sys

from ...auth import Permissions
from ...image_registry import ImageRegistry
from ...waiters import TimeoutError, wait_for
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def push(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    inspect_cmd = f"docker image inspect {args.local_image}"
    try:
        subprocess.run(
            shlex.split(inspect_cmd),
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError:
        print(
            f"Could not find locally built image '{args.local_image}'. Is the repository name and tag correct?",
            file=sys.stderr,
        )
        return

    image_registry = ImageRegistry(
        context.roboto_service_base_url,
        context.http,
    )
    parts = args.local_image.split(":")
    if len(parts) == 1:
        repo, tag = parts[0], "latest"
    elif len(parts) == 2:
        repo, tag = parts
    else:
        raise ValueError("Invalid image format. Expected '<repository>:<tag>'.")
    repository = image_registry.create_repository(repo, org_id=args.org)
    credentials = image_registry.get_temporary_credentials(
        repository["repository_uri"], Permissions.ReadWrite, org_id=args.org
    )
    login_cmd = f"docker login --username {credentials.username} --password-stdin {credentials.registry_url}"
    try:
        subprocess.run(
            shlex.split(login_cmd),
            capture_output=True,
            check=True,
            input=credentials.password,
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        print(
            "Failed to set Docker credentials for Roboto's image registry.",
            file=sys.stderr,
        )
        if exc.stdout:
            print(exc.stdout, file=sys.stderr)
        if exc.stderr:
            print(exc.stderr, file=sys.stderr)
        return

    image_uri = f"{repository['repository_uri']}:{tag}"
    tag_cmd = f"docker tag {args.local_image} {image_uri}"
    try:
        subprocess.run(
            shlex.split(tag_cmd),
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        print(
            f"Failed to tag local image '{args.local_image}' as '{image_uri}'.",
            file=sys.stderr,
        )
        if exc.stdout:
            print(exc.stdout, file=sys.stderr)
        if exc.stderr:
            print(exc.stderr, file=sys.stderr)
        return

    push_cmd = f"docker push {image_uri}"
    with subprocess.Popen(
        shlex.split(push_cmd),
        text=True,
    ) as push_proc:
        try:
            push_proc.wait()
        except KeyboardInterrupt:
            push_proc.kill()
            return

    print("Waiting for image to be available...")
    try:
        wait_for(
            image_registry.repository_contains_image,
            args=[repository["repository_name"], tag, args.org],
            interval=lambda iteration: min((2**iteration) / 2, 32),
        )
        print(
            f"Image pushed successfully! You can now use '{image_uri}' in your Roboto Actions."
        )
    except TimeoutError:
        print(
            "Image could not be confirmed as successfully pushed. Try pushing again in a few minutes."
        )
    except KeyboardInterrupt:
        print("")


def push_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)

    parser.add_argument(
        "local_image",
        action="store",
        help=(
            "Specify the local image to push, in the format '<repository>:<tag>'. "
            "If no tag is specified, 'latest' is assumed. "
            "Image must exist locally (i.e. 'docker images' must list it)."
        ),
    )


push_command = RobotoCommand(
    name="push",
    logic=push,
    setup_parser=push_parser,
    command_kwargs={
        "help": (
            "Push a local container image into Roboto's image registry. "
            "Requires Docker CLI."
        )
    },
)



import collections.abc
import datetime
from typing import Any, Optional

from ...query import QuerySpecification
from ...serde import pydantic_jsonable_dict
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .invocation_delegate import (
    InvocationDelegate,
)
from .invocation_record import (
    ActionProvenance,
    ExecutableProvenance,
    InvocationDataSource,
    InvocationRecord,
    InvocationStatus,
    InvocationStatusRecord,
    LogRecord,
    LogsLocation,
    SourceProvenance,
)


class Invocation:
    __invocation_delegate: InvocationDelegate
    __record: InvocationRecord

    @classmethod
    def from_id(
        cls,
        invocation_id: str,
        invocation_delegate: InvocationDelegate,
        org_id: Optional[str] = None,
    ) -> "Invocation":
        record = invocation_delegate.get_by_id(invocation_id, org_id)
        return cls(record, invocation_delegate)

    @classmethod
    def query(
        cls,
        query: QuerySpecification,
        invocation_delegate: InvocationDelegate,
        org_id: Optional[str] = None,
    ) -> collections.abc.Generator["Invocation", None, None]:
        known = set(InvocationRecord.__fields__.keys())
        actual = set()
        for field in query.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of Invocation"
                if plural
                else "is not a known attribute of Invocation"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        paginated_results = invocation_delegate.query_invocations(query, org_id=org_id)
        while True:
            for record in paginated_results.items:
                yield cls(record, invocation_delegate)
            if paginated_results.next_token:
                query.after = paginated_results.next_token
                paginated_results = invocation_delegate.query_invocations(
                    query, org_id=org_id
                )
            else:
                break

    def __init__(
        self, record: InvocationRecord, invocation_delegate: InvocationDelegate
    ) -> None:
        self.__invocation_delegate = invocation_delegate
        self.__record = record

    def __str__(self) -> str:
        return str(self.__record)

    @property
    def action(self) -> ActionProvenance:
        return self.__record.provenance.action

    @property
    def compute_requirements(self) -> ComputeRequirements:
        return self.__record.compute_requirements

    @property
    def container_parameters(self) -> ContainerParameters:
        return self.__record.container_parameters

    @property
    def created(self) -> datetime.datetime:
        return self.__record.created

    @property
    def current_status(self) -> InvocationStatus:
        sorted_status_records = sorted(
            self.__record.status, key=lambda s: s.status.value
        )
        return sorted_status_records[-1].status

    @property
    def data_source(self) -> InvocationDataSource:
        return self.__record.data_source

    @property
    def executable(self) -> ExecutableProvenance:
        return self.__record.provenance.executable

    @property
    def id(self) -> str:
        return self.__record.invocation_id

    @property
    def input_data(self) -> list[str]:
        return self.__record.input_data

    @property
    def logs(self) -> Optional[LogsLocation]:
        return self.__record.logs

    @property
    def org_id(self) -> str:
        return self.__record.org_id

    @property
    def parameter_values(self) -> dict[str, Any]:
        return self.__record.parameter_values

    @property
    def record(self) -> InvocationRecord:
        return self.__record

    @property
    def reached_terminal_status(self) -> bool:
        return any(
            status_record.status.is_terminal() for status_record in self.__record.status
        )

    @property
    def source(self) -> SourceProvenance:
        return self.__record.provenance.source

    @property
    def status_log(self) -> list[InvocationStatusRecord]:
        return self.__record.status

    def is_queued_for_scheduling(self) -> bool:
        """
        An invocation is queued for scheduling if it has not yet reached the "Scheduled" status
        and it is not "Deadly".
        """
        previously_scheduled_or_deadly = (
            InvocationStatus.Scheduled,
            InvocationStatus.Deadly,
        )
        return all(
            status_record.status not in previously_scheduled_or_deadly
            for status_record in self.__record.status
        )

    def get_logs(
        self, page_token: Optional[str] = None
    ) -> collections.abc.Generator[LogRecord, None, None]:
        paginated_results = self.__invocation_delegate.get_logs(
            self.id,
            page_token,
            self.org_id,
        )
        while True:
            for record in paginated_results.items:
                yield record
            if paginated_results.next_token:
                paginated_results = self.__invocation_delegate.get_logs(
                    self.id,
                    paginated_results.next_token,
                    self.org_id,
                )
            else:
                break

    def stream_logs(
        self, last_read: Optional[str] = None
    ) -> collections.abc.Generator[LogRecord, None, Optional[str]]:
        streamed_results = self.__invocation_delegate.stream_logs(
            self.id,
            last_read,
            self.org_id,
        )
        while True:
            for record in streamed_results.items:
                yield record
            if streamed_results.has_next and streamed_results.last_read:
                streamed_results = self.__invocation_delegate.stream_logs(
                    self.id,
                    streamed_results.last_read,
                    self.org_id,
                )
            else:
                break

        return streamed_results.last_read

    def set_logs_location(self, logs: LogsLocation) -> None:
        updated_record = self.__invocation_delegate.set_logs_location(
            self.__record, logs
        )
        self.__record = updated_record

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)

    def refresh(self) -> None:
        self.__record = self.__invocation_delegate.get_by_id(self.id, self.org_id)

    def update_status(
        self, next_status: InvocationStatus, detail: Optional[str] = "None"
    ) -> None:
        if next_status == InvocationStatus.Failed:
            # Heuristic: if this is the third time the invocation has failed, it is Deadly
            num_failures = len(
                [
                    status_record
                    for status_record in self.__record.status
                    if status_record.status == InvocationStatus.Failed
                ]
            )
            if num_failures >= 2:
                next_status = InvocationStatus.Deadly

        if not self.current_status.can_transition_to(next_status):
            raise ValueError(
                f"Cannot transition invocation from {self.current_status} to {next_status}"
            )

        updated_record = self.__invocation_delegate.update_invocation_status(
            self.__record, next_status, detail
        )
        self.__record = updated_record



#  Copyright (c) 2023 Roboto Technologies, Inc.

import datetime
import enum
from typing import Any, Optional

import pydantic


class Administrator(str, enum.Enum):
    # Other supported type would be "Customer"
    Roboto = "Roboto"


class StorageLocation(str, enum.Enum):
    # Other supported locations might be "GCP" or "Azure"
    S3 = "S3"


# https://www.google.com/search?q=pydantic.Field+default_factory+not+evaluated+in+parse_obj&rlz=1C5CHFA_enUS1054US1054&oq=pydantic.Field+default_factory+not+evaluated+in+parse_obj&aqs=chrome..69i57j33i160.6235j0j7&sourceid=chrome&ie=UTF-8
class DatasetRecord(pydantic.BaseModel):
    # Primary key, defined in CDK
    org_id: str  # partition key
    dataset_id: str  # sort key

    administrator: Administrator
    # Persisted as ISO 8601 string in UTC
    created: datetime.datetime
    created_by: str
    description: Optional[str]
    metadata: dict[str, Any] = pydantic.Field(default_factory=dict)
    # Persisted as ISO 8601 string in UTC
    modified: datetime.datetime
    modified_by: str
    storage_location: StorageLocation
    tags: list[str] = pydantic.Field(default_factory=list)
    roboto_record_version: int = 0  # A protected field, incremented on every update



import collections.abc
import dataclasses
import pathlib
from typing import Any, Optional

import pathspec

from ...auth import Permissions
from ...query import QuerySpecification
from ...serde import (
    git_paths_match,
    pydantic_jsonable_dict,
)
from ...updates import (
    MetadataChangeset,
    StrSequence,
    UpdateCondition,
)
from ..files import File, FileDelegate, FileTag
from ..files.progress import (
    TqdmProgressMonitorFactory,
)
from .delegate import (
    Credentials,
    DatasetDelegate,
    StorageLocation,
)
from .record import Administrator, DatasetRecord


@dataclasses.dataclass
class FileUploadInfo:
    bucket: str
    key: str


class Dataset:
    __dataset_delegate: DatasetDelegate
    __file_delegate: FileDelegate
    __record: DatasetRecord
    __temp_credentials: Optional[Credentials] = None

    @classmethod
    def create(
        cls,
        dataset_delegate: DatasetDelegate,
        file_delegate: FileDelegate,
        administrator: Administrator = Administrator.Roboto,
        storage_location: StorageLocation = StorageLocation.S3,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        tags: Optional[list[str]] = None,
        description: Optional[str] = None,
    ) -> "Dataset":
        record = dataset_delegate.create_dataset(
            administrator,
            metadata,
            storage_location,
            tags,
            org_id,
            created_by,
            description,
        )
        return cls(record, dataset_delegate, file_delegate)

    @classmethod
    def from_id(
        cls,
        dataset_id: str,
        dataset_delegate: DatasetDelegate,
        file_delegate: FileDelegate,
        org_id: Optional[str] = None,
    ) -> "Dataset":
        record = dataset_delegate.get_dataset_by_primary_key(dataset_id, org_id)
        return cls(record, dataset_delegate, file_delegate)

    @classmethod
    def query(
        cls,
        query: QuerySpecification,
        dataset_delegate: DatasetDelegate,
        file_delegate: FileDelegate,
        org_id: Optional[str] = None,
    ) -> collections.abc.Generator["Dataset", None, None]:
        known = set(DatasetRecord.__fields__.keys())
        actual = set()
        for field in query.fields():
            # Support dot notation for nested fields
            # E.g., "metadata.SoftwareVersion"
            if "." in field:
                actual.add(field.split(".")[0])
            else:
                actual.add(field)
        unknown = actual - known
        if unknown:
            plural = len(unknown) > 1
            msg = (
                "are not known attributes of Dataset"
                if plural
                else "is not a known attribute of Dataset"
            )
            raise ValueError(f"{unknown} {msg}. Known attributes: {known}")

        paginated_results = dataset_delegate.query_datasets(query, org_id=org_id)
        while True:
            for record in paginated_results.items:
                yield cls(record, dataset_delegate, file_delegate)
            if paginated_results.next_token:
                query.after = paginated_results.next_token
                paginated_results = dataset_delegate.query_datasets(
                    query, org_id=org_id
                )
            else:
                break

    def __init__(
        self,
        record: DatasetRecord,
        dataset_delegate: DatasetDelegate,
        file_delegate: FileDelegate,
    ) -> None:
        self.__dataset_delegate = dataset_delegate
        self.__file_delegate = file_delegate
        self.__record = record

    @property
    def dataset_id(self) -> str:
        return self.__record.dataset_id

    @property
    def metadata(self) -> dict[str, Any]:
        return self.__record.metadata.copy()

    @property
    def record(self) -> DatasetRecord:
        return self.__record

    @property
    def tags(self) -> list[str]:
        return self.__record.tags.copy()

    def complete_upload(self, upload_id: str) -> None:
        """
        Marks an upload as 'completed', which allows the Roboto Platform to evaluate triggers for automatic action on
        incoming data. This also aides reporting on partial upload failure cases.

        This API is called implicitly by the `upload_file` and `upload_directory` operations, and should only be
        used explicitly in power-user scenarios.

        Args:
            upload_id: An ID for a particular upload transaction which is granted by RobotoService when upload
                credentials are requested.

        Returns:
            None
        """
        self.__dataset_delegate.complete_upload(
            dataset_id=self.__record.dataset_id,
            upload_id=upload_id,
            org_id=self.__record.org_id,
        )

    def delete_dataset(self) -> None:
        self.__dataset_delegate.delete_dataset(self.__record)

    def delete_files(
        self,
        include_patterns: Optional[list[str]] = None,
        exclude_patterns: Optional[list[str]] = None,
    ) -> None:
        """
        Delete files associated with this dataset.

        `include_patterns` and `exclude_patterns` are lists of gitignore-style patterns.
        See https://git-scm.com/docs/gitignore.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.delete_files(
            ...    include_patterns=["**/*.png"],
            ...    exclude_patterns=["**/back_camera/**"]
            ... )

        """
        for file in self.list_files(include_patterns, exclude_patterns):
            file.delete()

    def download_files(
        self,
        out_path: pathlib.Path,
        include_patterns: Optional[list[str]] = None,
        exclude_patterns: Optional[list[str]] = None,
    ) -> None:
        """
        Download files associated with this dataset to the given directory.
        If `out_path` does not exist, it will be created.

        `include_patterns` and `exclude_patterns` are lists of gitignore-style patterns.
        See https://git-scm.com/docs/gitignore.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.download_files(
            ...     pathlib.Path("/tmp/tmp.nV1gdW5WHV"),
            ...     include_patterns=["**/*.g4"],
            ...     exclude_patterns=["**/test/**"]
            ... )
        """
        if (
            self.__record.storage_location != StorageLocation.S3
            or self.__record.administrator != Administrator.Roboto
        ):
            raise NotImplementedError(
                "Only S3-backed storage administered by Roboto is supported at this time."
            )

        if not out_path.is_dir():
            out_path.mkdir(parents=True)

        credentials = self.get_temporary_credentials(Permissions.ReadOnly)

        for file in self.list_files(include_patterns, exclude_patterns):
            local_path = out_path / file.relative_path
            file.download(
                local_path,
                credentials,
                progress_monitor_factory=TqdmProgressMonitorFactory(concurrency=1),
            )

    def get_temporary_credentials(
        self,
        permissions: Permissions = Permissions.ReadOnly,
        caller: Optional[str] = None,  # A Roboto user_id
        force_refresh: bool = False,
        transaction_id: Optional[str] = None,
    ) -> Credentials:
        if (
            force_refresh
            or permissions == Permissions.ReadWrite
            or self.__temp_credentials is None
            or self.__temp_credentials.is_expired()
        ):
            self.__temp_credentials = self.__dataset_delegate.get_temporary_credentials(
                self.__record, permissions, caller, transaction_id
            )
        return self.__temp_credentials

    def list_files(
        self,
        include_patterns: Optional[list[str]] = None,
        exclude_patterns: Optional[list[str]] = None,
    ) -> collections.abc.Generator[File, None, None]:
        """
        List files associated with this dataset.

        `include_patterns` and `exclude_patterns` are lists of gitignore-style patterns.
        See https://git-scm.com/docs/gitignore.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> for file in dataset.list_files(
            ...     include_patterns=["**/*.g4"],
            ...     exclude_patterns=["**/test/**"]
            ... ):
            ...     print(file.relative_path)
        """

        paginated_results = self.__dataset_delegate.list_files(
            self.__record.dataset_id, self.__record.org_id
        )
        while True:
            for record in paginated_results.items:
                file = File(record, self.__file_delegate)
                if not git_paths_match(
                    include_patterns=include_patterns,
                    exclude_patterns=exclude_patterns,
                    file=file.relative_path,
                ):
                    continue

                yield file
            if paginated_results.next_token:
                paginated_results = self.__dataset_delegate.list_files(
                    self.__record.dataset_id,
                    self.__record.org_id,
                    paginated_results.next_token,
                )
            else:
                break

    def put_metadata(
        self,
        metadata: dict[str, Any],
        updated_by: Optional[str] = None,  # A Roboto user_id
    ) -> None:
        """
        Set each `key`: `value` in this dict as dataset metadata if it doesn't exist, else overwrite the existing value.
        Keys must be strings. Dot notation is supported for nested keys.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.put_metadata({
            ...     "foo": "bar",
            ...     "baz.qux": 101,
            ... })

        """
        self.update(
            metadata_changeset=MetadataChangeset(put_fields=metadata),
            updated_by=updated_by,
        )

    def put_tags(
        self,
        tags: StrSequence,
        updated_by: Optional[str] = None,  # A Roboto user_id
    ) -> None:
        """Add each tag in this sequence if it doesn't exist"""
        self.update(
            metadata_changeset=MetadataChangeset(put_tags=tags),
            updated_by=updated_by,
        )

    def refresh(self) -> None:
        """Refresh the underlying dataset record with the latest server state."""
        self.__record = self.__dataset_delegate.get_dataset_by_primary_key(
            self.__record.dataset_id, self.__record.org_id
        )

    def remove_metadata(
        self,
        metadata: StrSequence,
        updated_by: Optional[str] = None,  # A Roboto user_id
    ) -> None:
        """
        Remove each key in this sequence from dataset metadata if it exists.
        Keys must be strings. Dot notation is supported for nested keys.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.remove_metadata(["foo", "baz.qux"])
        """
        self.update(
            metadata_changeset=MetadataChangeset(remove_fields=metadata),
            updated_by=updated_by,
        )

    def remove_tags(
        self,
        tags: StrSequence,
        updated_by: Optional[str] = None,  # A Roboto user_id
    ) -> None:
        """Remove each tag in this sequence if it exists"""
        self.update(
            metadata_changeset=MetadataChangeset(remove_tags=tags),
            updated_by=updated_by,
        )

    def to_dict(self) -> dict[str, Any]:
        return pydantic_jsonable_dict(self.__record)

    def upload_directory(
        self,
        directory_path: pathlib.Path,
        exclude_patterns: Optional[list[str]] = None,
    ) -> None:
        """
        Upload everything, recursively, in directory, ignoring files that match any of the ignore patterns.

        `exclude_patterns` is a list of gitignore-style patterns.
        See https://git-scm.com/docs/gitignore#_pattern_format.

        Example:
            >>> from roboto.domain import datasets
            >>> dataset = datasets.Dataset(...)
            >>> dataset.upload_directory(
            ...     pathlib.Path("/path/to/directory"),
            ...     exclude_patterns=[
            ...         "__pycache__/",
            ...         "*.pyc",
            ...         "node_modules/",
            ...         "**/*.log",
            ...     ],
            ... )
        """
        if not directory_path.is_dir():
            raise ValueError(f"{directory_path} is not a directory")

        exclude_spec: Optional[pathspec.PathSpec] = None
        if exclude_patterns is not None:
            exclude_spec = pathspec.GitIgnoreSpec.from_lines(
                exclude_patterns,
            )

        # Refresh the creds so we get access to the upload_id and all of the child upload_file requests succeed.
        credentials = self.get_temporary_credentials(
            Permissions.ReadWrite, force_refresh=True
        )

        def _upload_directory(directory_path: pathlib.Path, key_prefix: str) -> None:
            for path in directory_path.iterdir():
                if path.is_dir():
                    _upload_directory(path, f"{key_prefix}/{path.name}")
                else:
                    if exclude_spec is not None and exclude_spec.match_file(path):
                        continue
                    self.upload_file(
                        path, f"{key_prefix}/{path.name}", default_creds=credentials
                    )

        _upload_directory(directory_path, "")

        # Eventually we should error on credentials.upload_id is None, but we'll handle it gracefully
        # for backwards compatible rollout
        if credentials.upload_id is not None:
            self.complete_upload(upload_id=credentials.upload_id)

    def upload_file(
        self,
        local_file_path: pathlib.Path,
        key: str,
        default_creds: Optional[Credentials] = None,
    ) -> FileUploadInfo:
        """
        Uploads a file to the dataset's storage location.

        Args:
            local_file_path: The path to the file to upload.
            key: The key to use for the file in the dataset's storage location.
                It will be prefixed with the dataset's storage prefix.
            default_creds: The credentials to use for the upload. If not provided,
                temporary credentials will be requested.
        """
        if not local_file_path.is_file():
            raise ValueError(f"{local_file_path} is not a file")

        if (
            self.__record.storage_location != StorageLocation.S3
            or self.__record.administrator != Administrator.Roboto
        ):
            raise NotImplementedError(
                "Only S3-backed storage administered by Roboto is supported at this time."
            )

        credentials = (
            default_creds
            if default_creds is not None
            else self.get_temporary_credentials(
                Permissions.ReadWrite, force_refresh=True
            )
        )

        if credentials.upload_id is None:
            raise ValueError("No 'upload_id' brokered with credentials.")

        key = f"{credentials.required_prefix}/{key.lstrip('/')}"
        self.__file_delegate.upload_file(
            local_file_path,
            credentials.bucket,
            key,
            credentials,
            tags={
                FileTag.DatasetId: self.__record.dataset_id,
                FileTag.OrgId: self.__record.org_id,
                FileTag.CommonPrefix: credentials.required_prefix,
                FileTag.UploadId: credentials.upload_id,
            },
            progress_monitor_factory=TqdmProgressMonitorFactory(concurrency=1),
        )

        # Eventually we should error on credentials.upload_id is None, but we'll handle it gracefully
        # for backwards compatible rollout
        if default_creds is None and credentials.upload_id is not None:
            self.complete_upload(upload_id=credentials.upload_id)

        return FileUploadInfo(
            bucket=credentials.bucket,
            key=key,
        )

    def update(
        self,
        metadata_changeset: Optional[MetadataChangeset] = None,
        conditions: Optional[list[UpdateCondition]] = None,
        description: Optional[str] = None,
        updated_by: Optional[str] = None,  # A Roboto user_id
    ) -> None:
        updated = self.__dataset_delegate.update(
            self.__record,
            metadata_changeset=metadata_changeset,
            conditions=conditions,
            description=description,
            updated_by=updated_by,
        )
        self.__record = updated



#  Copyright (c) 2023 Roboto Technologies, Inc.


def validate_nonzero_gitpath_specs(value: list[str]) -> list[str]:
    filtered = list(filter(lambda p: p.strip() != "", value))

    if len(value) == 0:
        raise ValueError("Paths must not be empty.")
    elif len(filtered) == 0:
        raise ValueError(
            "Paths must have at least one entry which is not the empty string or just whitespace."
        )

    return value



#  Copyright (c) 2023 Roboto Technologies, Inc.

import abc
from typing import Any, Optional, Union

from roboto.sentinels import (
    NotSet,
    NotSetType,
)

from ...http import PaginatedList
from ...query import (
    ConditionType,
    QuerySpecification,
)
from ..actions import (
    ComputeRequirements,
    ContainerParameters,
)
from .trigger_record import (
    TriggerForEachPrimitive,
    TriggerRecord,
)


class TriggerDelegate(abc.ABC):
    @abc.abstractmethod
    def create_trigger(
        self,
        name: str,
        action_name: str,
        required_inputs: list[str],
        for_each: TriggerForEachPrimitive,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,  # A Roboto user_id
        compute_requirement_overrides: Optional[ComputeRequirements] = None,
        container_parameter_overrides: Optional[ContainerParameters] = None,
        condition: Optional[ConditionType] = None,
        additional_inputs: Optional[list[str]] = None,
        parameter_values: Optional[dict[str, Any]] = None,
        action_owner_id: Optional[str] = None,
        action_digest: Optional[str] = None,
    ) -> TriggerRecord:
        raise NotImplementedError("create_trigger")

    @abc.abstractmethod
    def get_trigger_by_primary_key(
        self, name: str, org_id: Optional[str] = None
    ) -> TriggerRecord:
        raise NotImplementedError("get_trigger_by_primary_key")

    @abc.abstractmethod
    def query_triggers(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[TriggerRecord]:
        raise NotImplementedError("query_triggers")

    @abc.abstractmethod
    def delete_trigger(self, name: str, org_id: str) -> None:
        raise NotImplementedError("delete_trigger")

    @abc.abstractmethod
    def update_trigger(
        self,
        name: str,
        org_id: str,
        updated_by: Optional[str] = None,
        action_name: Union[str, NotSetType] = NotSet,
        action_owner_id: Union[str, NotSetType] = NotSet,
        action_digest: Optional[Union[str, NotSetType]] = NotSet,
        required_inputs: Union[list[str], NotSetType] = NotSet,
        for_each: Union[TriggerForEachPrimitive, NotSetType] = NotSet,
        enabled: Union[bool, NotSetType] = NotSet,
        additional_inputs: Optional[Union[list[str], NotSetType]] = NotSet,
        compute_requirement_overrides: Optional[
            Union[ComputeRequirements, NotSetType]
        ] = NotSet,
        container_parameter_overrides: Optional[
            Union[ContainerParameters, NotSetType]
        ] = NotSet,
        condition: Optional[Union[ConditionType, NotSetType]] = NotSet,
        parameter_values: Optional[Union[dict[str, Any], NotSetType]] = NotSet,
    ) -> TriggerRecord:
        raise NotImplementedError("update_trigger")



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import time
from typing import Any, Optional

from ...domain.actions import Invocation
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def get_logs(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    invocation = Invocation.from_id(
        args.invocation_id,
        invocation_delegate=context.invocations,
        org_id=args.org,
    )

    for log_record in invocation.get_logs():
        print(log_record.log)


def stream_logs(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    invocation = Invocation.from_id(
        args.invocation_id,
        invocation_delegate=context.invocations,
        org_id=args.org,
    )

    last_read: Optional[Any] = None
    wait_msg = ""
    try:
        while True:
            try:
                log_record_generator = invocation.stream_logs(last_read)
                while True:
                    log_record = next(log_record_generator)
                    if wait_msg:
                        print("\r", " " * len(wait_msg), end="\r", flush=True)
                        wait_msg = ""
                    print(log_record.log)
            except StopIteration as stop:
                if invocation.reached_terminal_status:
                    break

                if not wait_msg:
                    wait_msg = "Waiting for logs..."
                    print(wait_msg, end="", flush=True)

                last_read = stop.value
                time.sleep(2)
                invocation.refresh()
    except KeyboardInterrupt:
        pass  # Swallow

    if wait_msg:
        print("\r", " " * len(wait_msg), end="\r", flush=True)


def main(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    if args.tail:
        stream_logs(args, context, parser)
    else:
        get_logs(args, context, parser)


def get_logs_parser(parser: argparse.ArgumentParser):
    parser.add_argument("invocation_id")
    parser.add_argument("--tail", required=False, action="store_true")
    add_org_arg(parser=parser)


get_logs_command = RobotoCommand(
    name="logs",
    logic=main,
    setup_parser=get_logs_parser,
    command_kwargs={"help": "Get invocation logs."},
)



#  Copyright (c) 2023 Roboto Technologies, Inc.

import abc
from typing import Any, Optional

from .user_record import UserRecord


class UserDelegate(abc.ABC):
    @abc.abstractmethod
    def get_user_by_id(self, user_id: Optional[str] = None) -> UserRecord:
        raise NotImplementedError("get_user_by_id")

    @abc.abstractmethod
    def delete_user(self, user_id: Optional[str] = None) -> None:
        raise NotImplementedError("delete_user")

    @abc.abstractmethod
    def update_user(self, user_id: str, updates: dict[str, Any]) -> UserRecord:
        raise NotImplementedError("update_user")



from typing import Any, Optional
import urllib.parse

from ...auth import Permissions
from ...exceptions import RobotoHttpExceptionParse
from ...http import (
    HttpClient,
    PaginatedList,
    roboto_headers,
)
from ...query import QuerySpecification
from ...serde import pydantic_jsonable_dict
from ...updates import (
    MetadataChangeset,
    UpdateCondition,
)
from ..files import FileRecord
from .delegate import Credentials, DatasetDelegate
from .http_resources import (
    CreateDatasetRequest,
    UpdateDatasetRequest,
)
from .record import (
    Administrator,
    DatasetRecord,
    StorageLocation,
)


class DatasetHttpDelegate(DatasetDelegate):
    __http_client: HttpClient
    __roboto_service_base_url: str

    def __init__(self, roboto_service_base_url: str, http_client: HttpClient) -> None:
        super().__init__()
        self.__http_client = http_client
        self.__roboto_service_base_url = roboto_service_base_url

    def headers(
        self, org_id: Optional[str] = None, user_id: Optional[str] = None
    ) -> dict[str, str]:
        return roboto_headers(
            org_id=org_id,
            user_id=user_id,
            additional_headers={"Content-Type": "application/json"},
        )

    def create_dataset(
        self,
        administrator: Administrator = Administrator.Roboto,
        metadata: Optional[dict[str, Any]] = None,
        storage_location: StorageLocation = StorageLocation.S3,
        tags: Optional[list[str]] = None,
        org_id: Optional[str] = None,
        created_by: Optional[str] = None,
        description: Optional[str] = None,
    ) -> DatasetRecord:
        """
        Create a new dataset.
        """
        url = f"{self.__roboto_service_base_url}/v1/datasets"
        request_body = CreateDatasetRequest(
            administrator=administrator,
            metadata=metadata if metadata is not None else {},
            storage_location=storage_location,
            description=description,
            tags=tags if tags is not None else [],
        )

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url,
                data=pydantic_jsonable_dict(request_body),
                headers=self.headers(org_id, created_by),
            )

        return DatasetRecord.parse_obj(response.from_json(json_path=["data"]))

    def delete_dataset(self, record: DatasetRecord) -> None:
        """
        Delete a dataset.
        """
        url = f"{self.__roboto_service_base_url}/v1/datasets/{record.dataset_id}"

        with RobotoHttpExceptionParse():
            self.__http_client.delete(
                url,
                headers=self.headers(record.org_id),
            )

    def get_dataset_by_primary_key(
        self,
        dataset_id: str,
        org_id: Optional[str] = None,
    ) -> DatasetRecord:
        """
        Get a dataset by its primary key (org_id, dataset_id)
        """
        url = f"{self.__roboto_service_base_url}/v1/datasets/{dataset_id}"

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url, headers=self.headers(org_id))

        return DatasetRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_temporary_credentials(
        self,
        record: DatasetRecord,
        permissions: Permissions,
        caller: Optional[str] = None,
        transaction_id: Optional[str] = None,
    ) -> Credentials:
        """
        Get temporary credentials to access a dataset.
        """
        query_params = {"mode": permissions.value}
        encoded_qs = urllib.parse.urlencode(query_params)
        url = f"{self.__roboto_service_base_url}/v1/datasets/{record.dataset_id}/credentials?{encoded_qs}"

        if transaction_id:
            url += f"&transaction_id={transaction_id}"

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url, self.headers(record.org_id, caller))

        return Credentials.parse_obj(response.from_json(json_path=["data"]))

    def complete_upload(
        self, dataset_id: str, upload_id: str, org_id: Optional[str] = None
    ) -> None:
        """
        Marks an upload as 'completed', which allows the Roboto Platform to evaluate triggers for automatic action on
        incoming data. This also aides reporting on partial upload failure cases.
        """
        url = f"{self.__roboto_service_base_url}/v1/datasets/{dataset_id}/uploads/{upload_id}/complete"

        with RobotoHttpExceptionParse():
            self.__http_client.put(url=url, headers=self.headers(org_id))

    def list_files(
        self,
        dataset_id: str,
        org_id: Optional[str] = None,
        page_token: Optional[str] = None,
    ) -> PaginatedList[FileRecord]:
        """
        List files associated with dataset.

        Files are associated with datasets in an eventually-consistent manner,
        so there will likely be delay between a file being uploaded and it appearing in this list.
        """
        url = f"{self.__roboto_service_base_url}/v1/datasets/{dataset_id}/files"
        if page_token:
            encoded_qs = urllib.parse.urlencode({"page_token": str(page_token)})
            url = f"{url}?{encoded_qs}"

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url, headers=self.headers(org_id))

        unmarshalled = response.from_json(json_path=["data"])
        return PaginatedList(
            items=[FileRecord.parse_obj(file) for file in unmarshalled["items"]],
            next_token=unmarshalled["next_token"],
        )

    def query_datasets(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[DatasetRecord]:
        url = f"{self.__roboto_service_base_url}/v1/datasets/query"
        post_body = pydantic_jsonable_dict(query, exclude_none=True)
        with RobotoHttpExceptionParse():
            res = self.__http_client.post(
                url,
                data=post_body,
                headers=self.headers(org_id),
            )

        unmarshalled = res.from_json(json_path=["data"])
        return PaginatedList(
            items=[
                DatasetRecord.parse_obj(dataset) for dataset in unmarshalled["items"]
            ],
            next_token=unmarshalled["next_token"],
        )

    def update(
        self,
        record: DatasetRecord,
        metadata_changeset: Optional[MetadataChangeset] = None,
        conditions: Optional[list[UpdateCondition]] = None,
        description: Optional[str] = None,
        updated_by: Optional[str] = None,
    ) -> DatasetRecord:
        url = f"{self.__roboto_service_base_url}/v1/datasets/{record.dataset_id}"
        payload = UpdateDatasetRequest(
            metadata_changeset=metadata_changeset,
            description=description,
            conditions=conditions,
        )
        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url,
                data=pydantic_jsonable_dict(payload, exclude_none=True),
                headers=self.headers(record.org_id, updated_by),
            )

        return DatasetRecord.parse_obj(response.from_json(json_path=["data"]))



#  Copyright (c) 2023 Roboto Technologies, Inc.

from typing import Any, Optional

from ...exceptions import RobotoHttpExceptionParse
from ...http import HttpClient, roboto_headers
from ...serde import pydantic_jsonable_dict
from .user_delegate import UserDelegate
from .user_http_resources import UpdateUserRequest
from .user_record import UserRecord


class UserHttpDelegate(UserDelegate):
    __http_client: HttpClient

    def __init__(self, http_client: HttpClient):
        super().__init__()
        self.__http_client = http_client

    def get_user_by_id(self, user_id: Optional[str] = None) -> UserRecord:
        url = self.__http_client.url("v1/users")
        headers = roboto_headers(user_id=user_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.get(url=url, headers=headers)

        return UserRecord.parse_obj(response.from_json(json_path=["data"]))

    def delete_user(self, user_id: Optional[str] = None) -> None:
        url = self.__http_client.url("v1/users")
        headers = roboto_headers(user_id=user_id)

        with RobotoHttpExceptionParse():
            self.__http_client.delete(url=url, headers=headers)

    def update_user(self, user_id: str, updates: dict[str, Any]) -> UserRecord:
        url = self.__http_client.url("v1/users")
        headers = roboto_headers(user_id=user_id)

        with RobotoHttpExceptionParse():
            response = self.__http_client.put(
                url=url,
                headers=headers,
                data=pydantic_jsonable_dict(UpdateUserRequest(updates=updates)),
            )

        return UserRecord.parse_obj(response.from_json(json_path=["data"]))



from typing import Any, Optional, Union

from ...exceptions import RobotoHttpExceptionParse
from ...http import (
    HttpClient,
    PaginatedList,
    roboto_headers,
)
from ...logging import default_logger
from ...query import QuerySpecification
from ...sentinels import (
    NotSet,
    NotSetType,
    is_set,
)
from ...serde import pydantic_jsonable_dict
from ...updates import MetadataChangeset
from .action_container_resources import (
    ComputeRequirements,
    ContainerParameters,
)
from .action_delegate import ActionDelegate
from .action_http_resources import (
    CreateActionRequest,
    SetActionAccessibilityRequest,
    UpdateActionRequest,
)
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionRecord,
    ActionReference,
)

logger = default_logger()


class ActionHttpDelegate(ActionDelegate):
    __http_client: HttpClient
    __roboto_service_base_url: str

    def __init__(self, roboto_service_base_url: str, http_client: HttpClient) -> None:
        super().__init__()
        self.__http_client = http_client
        self.__roboto_service_base_url = roboto_service_base_url

    def headers(
        self,
        org_id: Optional[str] = None,
        user_id: Optional[str] = None,
        resource_owner_id: Optional[str] = None,
    ) -> dict[str, str]:
        return roboto_headers(
            org_id=org_id,
            user_id=user_id,
            resource_owner_id=resource_owner_id,
            additional_headers={"Content-Type": "application/json"},
        )

    def create_action(
        self,
        name: str,
        parameters: Optional[list[ActionParameter]] = None,
        uri: Optional[str] = None,
        inherits: Optional[ActionReference] = None,
        compute_requirements: Optional[ComputeRequirements] = None,
        container_parameters: Optional[ContainerParameters] = None,
        description: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        tags: Optional[list[str]] = None,
        created_by: Optional[str] = None,  # A Roboto user_id
        org_id: Optional[str] = None,
    ) -> ActionRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions"
        request_body = CreateActionRequest(
            name=name,
            parameters=parameters,
            description=description,
            uri=uri,
            inherits=inherits,
            metadata=metadata,
            tags=tags,
            compute_requirements=compute_requirements,
            container_parameters=container_parameters,
        )

        with RobotoHttpExceptionParse():
            response = self.__http_client.post(
                url,
                data=pydantic_jsonable_dict(request_body, exclude_none=True),
                headers=self.headers(org_id, created_by),
            )

        return ActionRecord.parse_obj(response.from_json(json_path=["data"]))

    def get_action_by_primary_key(
        self,
        name: str,
        org_id: Optional[str] = None,
        digest: Optional[str] = None,
        action_owner_id: Optional[str] = None,
    ) -> ActionRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions/{name}"

        if digest:
            url += f"?digest={digest}"

        with RobotoHttpExceptionParse():
            res = self.__http_client.get(
                url,
                headers=self.headers(
                    org_id=org_id,
                    resource_owner_id=action_owner_id,
                ),
            )

        return ActionRecord.parse_obj(res.from_json(json_path=["data"]))

    def delete_action(self, record: ActionRecord) -> None:
        url = f"{self.__roboto_service_base_url}/v1/actions/{record.name}"
        with RobotoHttpExceptionParse():
            self.__http_client.delete(url, headers=self.headers(record.org_id))

    def query_actions(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[ActionRecord]:
        url = f"{self.__roboto_service_base_url}/v1/actions/query"
        post_body = pydantic_jsonable_dict(query, exclude_none=True)
        with RobotoHttpExceptionParse():
            res = self.__http_client.post(
                url,
                data=post_body,
                headers=self.headers(org_id),
            )
        unmarshalled = res.from_json(json_path=["data"])
        return PaginatedList(
            items=[
                ActionRecord.parse_obj(dataset) for dataset in unmarshalled["items"]
            ],
            next_token=unmarshalled["next_token"],
        )

    def query_actions_on_action_hub(
        self,
        query: QuerySpecification,
        org_id: Optional[str] = None,
    ) -> PaginatedList[ActionRecord]:
        url = f"{self.__roboto_service_base_url}/v1/actions/query/actionhub"
        post_body = pydantic_jsonable_dict(query, exclude_none=True)
        with RobotoHttpExceptionParse():
            res = self.__http_client.post(
                url,
                data=post_body,
                headers=self.headers(org_id),
            )
        unmarshalled = res.from_json(json_path=["data"])
        return PaginatedList(
            items=[
                ActionRecord.parse_obj(dataset) for dataset in unmarshalled["items"]
            ],
            next_token=unmarshalled["next_token"],
        )

    def set_accessibility(
        self,
        record: ActionRecord,
        accessibility: Accessibility,
    ) -> ActionRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions/{record.name}/accessibility"
        request_body = SetActionAccessibilityRequest(accessibility=accessibility)

        with RobotoHttpExceptionParse():
            res = self.__http_client.put(
                url,
                data=pydantic_jsonable_dict(request_body),
                headers=self.headers(record.org_id),
            )

        return ActionRecord.parse_obj(res.from_json(json_path=["data"]))

    def update(
        self,
        record: ActionRecord,
        compute_requirements: Union[ComputeRequirements, NotSetType] = NotSet,
        container_parameters: Union[ContainerParameters, NotSetType] = NotSet,
        description: Optional[Union[str, NotSetType]] = NotSet,
        metadata_changeset: Union[MetadataChangeset, NotSetType] = NotSet,
        parameter_changeset: Union[ActionParameterChangeset, NotSetType] = NotSet,
        uri: Union[str, NotSetType] = NotSet,
        updated_by: Optional[str] = None,
    ) -> ActionRecord:
        url = f"{self.__roboto_service_base_url}/v1/actions/{record.name}"

        updates: dict[str, Any] = dict()
        if is_set(compute_requirements):
            updates["compute_requirements"] = compute_requirements
        if is_set(container_parameters):
            updates["container_parameters"] = container_parameters
        if is_set(description):
            updates["description"] = description
        if is_set(metadata_changeset):
            updates["metadata_changeset"] = metadata_changeset
        if is_set(parameter_changeset):
            updates["parameter_changeset"] = parameter_changeset
        if is_set(uri):
            updates["uri"] = uri

        request_body = UpdateActionRequest.parse_obj(updates)

        with RobotoHttpExceptionParse():
            res = self.__http_client.put(
                url,
                data=pydantic_jsonable_dict(request_body, exclude_unset=True),
                headers=self.headers(record.org_id, updated_by),
            )

        return ActionRecord.parse_obj(res.from_json(json_path=["data"]))



#  Copyright (c) 2023 Roboto Technologies, Inc.

"""
SerDe: Serialization + Deserialization

This module contains utility functions for common serialization and deserialization scenarios used
throughout the Roboto python codebase.
"""

from .dicts import (
    case_insensitive_get,
    pydantic_jsonable_dict,
    pydantic_jsonable_dict_map,
    pydantic_jsonable_dicts,
    safe_dict_drill,
)
from .paths import (
    git_paths_match,
    git_paths_to_spec,
)

__all__ = [
    "pydantic_jsonable_dict",
    "pydantic_jsonable_dicts",
    "pydantic_jsonable_dict_map",
    "safe_dict_drill",
    "case_insensitive_get",
    "git_paths_to_spec",
    "git_paths_match",
]



#  Copyright (c) 2023 Roboto Technologies, Inc.
import argparse
import collections.abc
import enum
import json
import shlex
import typing

import pydantic

from ...domain import actions
from ..command import KeyValuePairsAction
from .exceptions import ParseError


class ActionReferenceParser(argparse.Action):
    """Parse an action reference string into its org (if given) and action name components."""

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: typing.Union[str, collections.abc.Sequence[str], None],
        option_string: typing.Optional[str] = None,
    ):
        if not isinstance(values, str):
            raise ValueError("ActionReference must be a string")

        org_name_parts = values.split("/", maxsplit=1)
        name_part = org_name_parts[-1]
        name_digest_parts = name_part.split("@", maxsplit=1)

        parse_result = dict()

        if len(org_name_parts) == 2:
            # Org is explicit: Action belongs to the org specified in the reference
            # Example: "some-org/an-action"
            action_owner = org_name_parts[0]
            parse_result["owner"] = action_owner

        if len(name_digest_parts) == 2:
            action_name, action_digest = name_digest_parts
            parse_result["name"] = action_name
            parse_result["digest"] = action_digest
        elif len(name_digest_parts) == 1:
            action_name = name_digest_parts[0]
            parse_result["name"] = action_name

        setattr(namespace, self.dest, actions.ActionReference.parse_obj(parse_result))


def add_action_reference_arg(
    parser: argparse.ArgumentParser,
    positional: bool = True,
    required: bool = True,
    arg_help: str = "Partially or fully qualified reference to Action on which to operate.",
    arg_name="action",
) -> None:
    arg = arg_name if positional else f"--{arg_name}"
    kwargs: dict[str, typing.Any] = {
        "metavar": "<action_name> | <action_name>@<digest> | <org>/<action_name> | <org>/<action_name>@<digest>",  # noqa: E501
        "action": ActionReferenceParser,
        "help": arg_help,
    }
    if not positional and required:
        kwargs["required"] = True

    parser.add_argument(
        arg,
        **kwargs,
    )


class ActionParameterArg(argparse.Action):
    """Parse an action parameter string an ActionParameter value object."""

    METAVAR: typing.ClassVar[
        str
    ] = "'name=<name>|required=<true|false>|default=<value>|description=<description>'"  # noqa: E501

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: typing.Union[str, collections.abc.Sequence[str], None],
        option_string: typing.Optional[str] = None,
    ):
        if not values:
            return

        parameters = []
        for parameter_str in values:
            parts = parameter_str.split("|")
            parse_result = {}
            for part in parts:
                trimmed = part.strip()
                if not trimmed:
                    continue

                parts = trimmed.split("=", maxsplit=1)
                if not len(parts) == 2:
                    raise argparse.ArgumentError(
                        self,
                        f"Unable to parse action parameter '{parameter_str}'. "
                        f"Parameter '{part}' is not parsable as a <key>=<value> pair.",
                    )

                key = parts[0].strip()
                value = parts[1].strip()
                parsed_value: typing.Any = value
                if key == "required":
                    parsed_value = parsed_value.lower() == "true"
                else:
                    try:
                        parsed_value = json.loads(parsed_value)
                    except Exception:
                        pass
                parse_result[key] = parsed_value

            try:
                action_parameter = actions.ActionParameter.parse_obj(parse_result)
            except pydantic.ValidationError as exc:
                raise argparse.ArgumentError(
                    self,
                    f"Unable to parse action parameter '{parameter_str}'. Reason: {exc}",
                ) from None
            parameters.append(action_parameter)

        existing_parameters = getattr(namespace, self.dest, None)
        if not existing_parameters:
            existing_parameters = []
        setattr(namespace, self.dest, [*existing_parameters, *parameters])


class DockerInstructionForm(enum.Enum):
    """The form of a CMD instruction."""

    Exec = "exec"
    Shell = "shell"


null = object()


def parse_compute_requirements(
    args: argparse.Namespace,
    defaults: typing.Optional[actions.ComputeRequirements] = None,
) -> typing.Optional[actions.ComputeRequirements]:
    if args.vcpu is None and args.memory is None and args.storage is None:
        return None

    defaults = defaults if defaults else actions.ComputeRequirements()

    try:
        kwargs = {
            key: value
            for key, value in [
                ("vCPU", args.vcpu if args.vcpu else defaults.vCPU),
                ("memory", args.memory if args.memory else defaults.memory),
                ("storage", args.storage if args.storage else defaults.storage),
            ]
            if value is not None
        }
        if not kwargs:
            return None
        return actions.ComputeRequirements.parse_obj(kwargs)
    except pydantic.ValidationError as exc:
        for err in exc.errors():
            err_msg = err.get("msg")
            msg = err_msg if err_msg else err
            raise ParseError(msg) from None

    return None


def add_compute_requirements_args(parser: argparse.ArgumentParser) -> None:
    resource_requirements_group = parser.add_argument_group(
        "Resource requirements",
        "Specify required compute resources.",
    )
    resource_requirements_group.add_argument(
        "--vcpu",
        required=False,
        type=int,
        choices=[256, 512, 1024, 2048, 4096, 8192, 16384],
        help="CPU units to dedicate to action invocation. Defaults to 512 (0.5vCPU).",
    )

    resource_requirements_group.add_argument(
        "--memory",
        required=False,
        type=int,
        help=(
            "Memory (in MiB) to dedicate to action invocation. Defaults to 1024 (1 GiB). "
            "Supported values range from 512 (0.5 GiB) to 122880 (120 GiB). "
            "Supported values are tied to selected vCPU resources. See documentation for more information."
        ),
    )

    resource_requirements_group.add_argument(
        "--storage",
        required=False,
        type=int,
        help=(
            "Ephemeral storage (in GiB) to dedicate to action invocation. Defaults to 21 GiB. "
            "Supported values range from 21 to 200, inclusive."
        ),
    )

    # Placeholder
    resource_requirements_group.add_argument(
        "--gpu",
        required=False,
        default=False,
        action="store_true",
        help=(
            "This is a placeholder; it currently does nothing. "
            "In the future, setting this option will invoke the action in a GPU-enabled compute environment."
        ),
    )


def parse_container_overrides(
    args: argparse.Namespace,
    defaults: typing.Optional[actions.ContainerParameters] = None,
) -> typing.Optional[actions.ContainerParameters]:
    if (
        args.entry_point is None
        and args.command is None
        and args.workdir is None
        and args.env is None
    ):
        return None

    defaults = defaults if defaults else actions.ContainerParameters()

    try:
        entry_point: typing.Union[list[str], object] = defaults.entry_point
        if args.entry_point is null:
            entry_point = null
        elif args.entry_point is not None:
            entry_point = [args.entry_point]

        command: typing.Union[list[str], object] = defaults.command
        if args.command is null:
            command = null
        elif args.command is not None:
            command_form = DockerInstructionForm(args.command_form)
            command = []
            if command_form == DockerInstructionForm.Exec and len(args.command):
                lexxer = shlex.shlex(args.command, posix=True, punctuation_chars=True)
                lexxer.whitespace_split = True
                command = list(lexxer)
            else:
                command = [args.command]

        kwargs = {
            key: value
            for key, value in [
                ("entry_point", entry_point),
                ("command", command),
                ("workdir", args.workdir if args.workdir else defaults.workdir),
                ("env_vars", args.env if args.env else defaults.env_vars),
            ]
            if value is not None
        }
        if not kwargs:
            return None
        return actions.ContainerParameters.parse_obj(
            {key: value if value is not null else None for key, value in kwargs.items()}
        )
    except pydantic.ValidationError as exc:
        for err in exc.errors():
            err_msg = err.get("msg")
            msg = err_msg if err_msg else err
            raise ParseError(msg) from None

    return None


def add_container_parameters_args(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group(
        "Container parameters",
        "Specify parameters to pass to the action's Docker container at runtime.",
    )

    group.add_argument(
        "--entrypoint",
        required=False,
        type=lambda s: s if s != "null" else null,
        dest="entry_point",
        help=(
            "Container ENTRYPOINT override."
            ' Supports passing empty string ("") as an override, which unsets the ENTRYPOINT specified in the docker image.'  # noqa: E501
            " If updating or invoking action which has existing ENTRYPOINT override, specify 'null' to remove the override."  # noqa: E501
            " Refer to docker documentation for more: "
            "https://docs.docker.com/engine/reference/builder/#entrypoint"
            " and https://docs.docker.com/engine/reference/run/#entrypoint-default-command-to-execute-at-runtime"
        ),
    )

    group.add_argument(
        "--command",
        required=False,
        type=lambda s: s if s != "null" else null,
        dest="command",
        help=(
            "Container CMD override."
            " If updating or invoking action which has existing CMD override, specify 'null' to remove the override."
            " Refer to docker documentation for more: "
            "https://docs.docker.com/engine/reference/builder/#cmd and"
            " https://docs.docker.com/engine/reference/run/#cmd-default-command-or-options"
        ),
    )

    group.add_argument(
        "--command-form",
        required=False,
        choices=[form.value for form in DockerInstructionForm],
        default=DockerInstructionForm.Exec.value,
        dest="command_form",
        help=(
            "In 'exec' form, the provided '--command' str is split into a list of strings"
            ' (e.g., \'--command "-c \'print(123)\'"\' is parsed as ["-c", "print(123)"]).'
            " In 'shell' form, the provided '--command' str is not split"
            " (e.g., '--command \"python -c 'print(123)'\"' is parsed as [\"python -c 'print(123)'\"])."
        ),
    )

    group.add_argument(
        "--workdir",
        required=False,
        type=lambda s: s if s != "null" else null,
        dest="workdir",
        help=(
            "If updating, specify 'null' to clear existing workdir."
            " Refer to docker documentation for more: https://docs.docker.com/engine/reference/run/#workdir"
        ),
    )

    group.add_argument(
        "--env",
        required=False,
        metavar="KEY=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more 'key=value' formatted pairs to set as container ENV vars. "
            "Do not use ENV vars for secrets (such as API keys). "
            "See documentation: https://docs.docker.com/engine/reference/run/#env-environment-variables"
        ),
    )



from typing import Optional

from ...http import PaginatedList
from .delegate import CommentDelegate
from .record import CommentRecord, EntityType


class Comment:
    __comment_delegate: CommentDelegate
    __record: CommentRecord

    @classmethod
    def create(
        cls,
        comment_text: str,
        entity_id: str,
        entity_type: EntityType,
        comment_delegate: CommentDelegate,
        created_by: Optional[str] = None,
        org_id: Optional[str] = None,
    ) -> "Comment":
        record = comment_delegate.create_comment(
            entity_type,
            entity_id,
            comment_text,
            org_id,
            created_by,
        )
        return cls(record, comment_delegate)

    @classmethod
    def from_id(
        cls,
        comment_id: str,
        comment_delegate: CommentDelegate,
        org_id: Optional[str] = None,
    ) -> "Comment":
        record = comment_delegate.get_comment_by_id(comment_id, org_id)
        return cls(record, comment_delegate)

    @classmethod
    def for_entity(
        cls,
        comment_delegate: CommentDelegate,
        entity_type: EntityType,
        entity_id: str,
        org_id: Optional[str] = None,
        page_token: Optional[str] = None,
    ) -> PaginatedList[CommentRecord]:
        return comment_delegate.get_comments_by_entity(
            entity_type, entity_id, org_id, page_token
        )

    def __init__(
        self,
        record: CommentRecord,
        comment_delegate: CommentDelegate,
    ) -> None:
        self.__comment_delegate = comment_delegate
        self.__record = record

    @property
    def record(self) -> CommentRecord:
        return self.__record

    def delete_comment(self) -> None:
        self.__comment_delegate.delete_comment(self.__record)

    def update_comment(self, comment_text: str) -> None:
        self.__record = self.__comment_delegate.update_comment(
            self.__record, comment_text
        )



import datetime
import urllib.parse

import pydantic


class FileRecord(pydantic.BaseModel):
    association_id: str  # e.g. dataset_id, collection_id, etc.; GSI PK of "association_id" index.
    file_id: str  # Table PK
    modified: datetime.datetime  # Persisted as ISO 8601 string in UTC
    relative_path: str  # path relative to some common prefix. Used as local path when downloaded.
    size: int  # bytes
    org_id: str
    uri: str  # GSI PK of "uri" index; GSI SK of "association_id" index
    upload_id: str = "NO_ID"  # Defaulted for backwards compatability

    @property
    def bucket(self) -> str:
        parsed_uri = urllib.parse.urlparse(self.uri)
        return parsed_uri.netloc

    @property
    def key(self) -> str:
        parsed_uri = urllib.parse.urlparse(self.uri)
        return parsed_uri.path.lstrip("/")



#  Copyright (c) 2023 Roboto Technologies, Inc.

import datetime
from enum import Enum

import pydantic


class EntityType(str, Enum):
    Dataset = "dataset"


class CommentRecord(pydantic.BaseModel):
    # Primary key, defined in CDK
    org_id: str  # partition key
    comment_id: str
    entity_type: EntityType
    entity_id: str
    # Persisted as ISO 8601 string in UTC
    created: datetime.datetime
    created_by: str
    comment_text: str
    mentions: list[str] = pydantic.Field(default_factory=list)
    # Persisted as ISO 8601 string in UTC
    modified: datetime.datetime
    modified_by: str

    # for v2
    # parent_comment_id: str
